"""Type stubs for the crypto C extension module."""

import numpy.typing as npt
import numpy as np

def encrypt_bytes(buf: bytearray, key: npt.NDArray[np.uint8]) -> None: ...
def decrypt_bytes(buf: bytearray, key: npt.NDArray[np.uint8]) -> None: ...
def encrypt_blocks(data: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], full_word_count: int) -> None: ...
def decrypt_blocks(data: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], full_word_count: int) -> None: ...
def encrypt_block(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], n: int) -> None: ...
def decrypt_block(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32], n: int) -> None: ...
def encrypt_block_fixed(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32]) -> None: ...
def decrypt_block_fixed(v: npt.NDArray[np.uint32], key: npt.NDArray[np.uint32]) -> None: ...
